# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .message_types import MessageType
from .return_codes import ReturnCode


class MethodResult:
    """Class representing the result of a method call."""

    message_type: MessageType
    return_code: ReturnCode
    payload: bytes

    def __init__(self):
        self.message_type = MessageType.RESPONSE
        self.return_code = ReturnCode.E_OK
        self.payload = b""
